/* -*- c++ -*- */
/*
 * Copyright 2004 Free Software Foundation, Inc.
 * 
 * This file is part of GNU Radio
 * 
 * GNU Radio is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * GNU Radio is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Radio; see the file COPYING.  If not, write to
 * the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * config.h is generated by configure.  It contains the results
 * of probing for features, options etc.  It should be the first
 * file included in your .cc file.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef DEBUG
#define DBG(x) x 
#else
#define DBG(x)
#endif

#include <gr_rds_data_decoder.h>
#include <gr_io_signature.h>

gr_rds_data_decoder_sptr gr_rds_make_data_decoder (gr_msg_queue_sptr msgq) {
  return gr_rds_data_decoder_sptr (new gr_rds_data_decoder (msgq));
}

static const int MIN_IN = 1;	// mininum number of input streams
static const int MAX_IN = 1;	// maximum number of input streams

gr_rds_data_decoder::gr_rds_data_decoder (gr_msg_queue_sptr msgq)
  : gr_sync_block ("gr_rds_data_decoder",
  gr_make_io_signature (MIN_IN, MAX_IN, sizeof (bool)),
  gr_make_io_signature (0, 0, 0))
{
	d_msgq=msgq;
	set_output_multiple(104); 	// Since a RDS datagroup contains 104 bits
	reset();
}

gr_rds_data_decoder::~gr_rds_data_decoder () {
}

void gr_rds_data_decoder::reset() {
	bit_counter=0;
	reg=0;
	reset_rds_data();
	enter_no_sync();
}

char gr_rds_data_decoder::transform_char(char z) {
	int row, col;
	static char transform_table[16][8]  = 
						{{' ', '0', '@', 'P', ' ', 'p', 'a', 'a'},
						{'!', '1', 'A', 'Q', 'a', 'q', 'a', 'a'},
						{'"', '2', 'B', 'R', 'b', 'r', 'e', 'e'},
						{'#', '3', 'C', 'S', 'c', 's', 'e', 'e'},
						{'�', '4', 'D', 'T', 'd', 't', 'i', 'i'},
						{'%', '5', 'E', 'U', 'e', 'u', 'i', 'i'},
						{'&', '6', 'F', 'V', 'f', 'v', 'o', 'o'},
						{'\'', '7', 'G', 'W', 'g', 'w', 'o', 'o'},
						{'(', '8', 'H', 'X', 'h', 'x', 'u', 'u'},
						{')', '9', 'I', 'Y', 'i', 'y', 'u', 'u'},
						{'*', ':', 'J', 'Z', 'j', 'z', 'N', 'n'},
						{'+', ';', 'K', '[', 'k', '{', 'c', 'c'},
						{',', '<', 'L', '\\', 'l', '|', 's', 's'},
						{'-', '=', 'M', ']', 'm', '}', 'b', 'g'},
						{'.', '>', 'N', '-', 'n', '-', 'i', 'i'},
						{'/', '?', 'O', '-', 'o', ' ', 'I', 'i'}};

	row = (z & 0x0F);
	col = ((z >> 4) & 0x0F);
	col -= 2; //FIXME
	return transform_table[row][col];
}

void gr_rds_data_decoder::reset_rds_data() {
	memset(radiotext,' ',sizeof(radiotext));
	radiotext[0][64] = radiotext[1][64] = '\0';
	radiotext_AB_flag=0;
  	traffic_programme=false;
  	traffic_announcement=false;
  	music_speech=false;
  	programme_type=0;
  	pi_country_identification=0;
    pi_area_coverage=0;
    pi_programme_reference_number=0;
    memset(programme_service_name,' ',sizeof(programme_service_name));
  	programme_service_name[9]='\0';
  	mono_stereo=false;
  	artificial_head=false;
  	compressed=false;
  	static_pty=false;  	
}

void gr_rds_data_decoder::enter_no_sync() {
	presync=false;
	d_state = ST_NO_SYNC;
	send_message(6,"");
}

void gr_rds_data_decoder::enter_sync(unsigned int sync_block_number) {
	wrong_blocks_counter=0;
	blocks_counter=0;
	block_bit_counter=0;
	block_number=(sync_block_number+1) % 4;
	group_assembly_started=false;
	DBG(printf("Starting Block Decode @ block %u\n",block_number);)
	d_state = ST_SYNC;
}

unsigned long gr_rds_data_decoder::calc_syndrome(unsigned long message, unsigned char mlen, unsigned long poly, unsigned char plen) {
	unsigned long reg=0;
	unsigned int i;

	for (i=mlen;i>0;i--)  {
		reg=(reg<<1) | ((message>>(i-1)) & 0x01);
		if (reg & (1<<plen)) reg=reg^poly;
	}
	for (i=plen;i>0;i--) {
		reg=reg<<1;
		if (reg & (1<<plen)) reg=reg^poly;
	}
	return (reg & ((1<<plen)-1));
}

void gr_rds_data_decoder::printbin(unsigned long number,unsigned char bits) {
	unsigned int i;
	for (i=bits;i>0;i--) putchar((number & (1L<<(i-1)))?'1':'0');
	putchar('\n');
}

unsigned long gr_rds_data_decoder::bin2dec(char *string) {
	unsigned int i;
	unsigned long reg=0;

	for (i=0;i<strlen(string);i++) {
		reg=(reg<<1) | ((string[i]=='0')?0:1);
	}	
	return reg;
}

void gr_rds_data_decoder::send_message(long msgtype, std::string msgtext) {
  if (!d_msgq->full_p()) {
		gr_message_sptr msg = gr_make_message_from_string(msgtext,msgtype,0,0);
		d_msgq->insert_tail(msg);
		msg.reset();
	}
}

void gr_rds_data_decoder::decode_pi(unsigned int program_information_code) {
	char pistring[5];

	DBG(printf("PI Code :%04X\n",program_information_code);)
	sprintf(pistring,"%04X",program_information_code);
	send_message(0,pistring);
	//TODO feed pi into pi_country_identification,pi_area_coverage and pi_programme_reference_number
}

void gr_rds_data_decoder::do_radiotext(unsigned int *group, bool is_version_A) {
	unsigned char new_radiotext_AB_flag=(group[1]>>4) & 0x01;
	unsigned char text_segment_address_code=group[1] & 0x0f;
	if (radiotext_AB_flag!=new_radiotext_AB_flag) {
		DBG(printf("Text A/B Flag Toggled!!!\n");)
		send_message(4,radiotext[0]);
		send_message(5,radiotext[1]);
		for(int i= 0; i < 64; i++) radiotext[0][i] = radiotext[1][i] = ' ';
		radiotext[0][64] = radiotext[1][64] = '\0';
		radiotext_AB_flag=new_radiotext_AB_flag;
	}
	DBG(printf("Text Segment Address Code: %u (Type: %c)\n",text_segment_address_code,is_version_A?'A':'B');)
	if (is_version_A) {
		radiotext[0][text_segment_address_code*4]=transform_char((group[2]>>8)&0xff);
		radiotext[0][text_segment_address_code*4+1]=transform_char(group[2]&0xff);
		radiotext[0][text_segment_address_code*4+2]=transform_char((group[3]>>8)&0xff);		
		radiotext[0][text_segment_address_code*4+3]=transform_char(group[3]&0xff);		
	}
	else {
		radiotext[0][text_segment_address_code*2]=transform_char((group[3]>>8)&0xff);
		radiotext[0][text_segment_address_code*2+1]=transform_char(group[3]&0xff);		
	}
	DBG(printf("Radio Text A: %s\n",radiotext[0]);)
	DBG(printf("Radio Text B: %s\n",radiotext[1]);)	
}

void gr_rds_data_decoder::do_type0(unsigned int *group, bool is_version_A) {
	unsigned char alternative_frequency_code_1;
	unsigned char alternative_frequency_code_2;
	char flagstring[8]="0000000";
	
	traffic_programme=(group[1]>>10) & 0x01;
	programme_type=(group[1]>>5) & 0x0F;
	traffic_announcement=(group[1]>>4) & 0x01;
	music_speech=(group[1]>>3) & 0x01;
	bool decoder_control_bit = (group[1]>>2) & 0x01;
	unsigned char segment_address = group[1] & 0x03;
	programme_service_name[segment_address*2]=transform_char((group[3]>>8)&0xff);
	programme_service_name[segment_address*2+1]=transform_char(group[3]&0xff);
	switch (segment_address) {
		case 0:
			mono_stereo=decoder_control_bit;
		break;
		case 1:
			artificial_head=decoder_control_bit;
		break;
		case 2:
			compressed=decoder_control_bit;
		break;
		case 3:
			static_pty=decoder_control_bit;
		break;
		default:
		break;
	}
	if (is_version_A) {
		alternative_frequency_code_1=(group[2]>>8)&0xff;
		alternative_frequency_code_2=group[2]&0xff;
		//TODO decode AFs		
	}
	else {
	}
	std::cout << "==>" << programme_service_name << "<== ";
	std::cout << '-' << (traffic_programme?"TP":"  ") << '-' << (traffic_announcement?"TA":"  ") ;
	std::cout << '-' << (music_speech?"Music":"Speech") << '-' << (mono_stereo?"MONO":"STEREO"); 
	std::cout << '-' << pty_table[programme_type] << std::endl;
	//std::cout << (alternative_frequency_code_1<205?alternative_frequency_code_1+875:alternative_frequency_code_1);
	//std::cout << '-' << (alternative_frequency_code_2<205?alternative_frequency_code_2+875:alternative_frequency_code_2) << std::endl;
	send_message(1,programme_service_name);
	send_message(2,pty_table[programme_type]);
	flagstring[0]=traffic_programme?'1':'0';	
	flagstring[1]=traffic_announcement?'1':'0';
	flagstring[2]=music_speech?'1':'0';
	flagstring[3]=mono_stereo?'1':'0';
	flagstring[4]=artificial_head?'1':'0';
	flagstring[5]=compressed?'1':'0';
	flagstring[6]=static_pty?'1':'0';
        send_message(3,flagstring);
	
}

void gr_rds_data_decoder::decode_group_by_type(unsigned char group_type_code, unsigned char version_code, unsigned int *group) {
	DBG(printf("Group Type Code:%u\n",group_type_code);)
	DBG(printf("Version:%c\n",(version_code==0)?'A':'B');)
	std::string message("");
	switch (version_code) {
		case 0:
			switch (group_type_code) {
				case 0:
					message="Basic tuning and switching information only (see 3.1.5.1)";
					do_type0(group,true);
				break;
				case 1:
					message="Programme Item Number and slow labelling codes only (see 3.1.5.2)";
				break;
				case 2:
					message="RadioText only (see 3.1.5.3)";
					do_radiotext(group,true);
				break;
				case 3:
					message="Applications Identification for ODA only (see 3.1.5.5)";
				break;
				case 4:
					message="Clock-time and date only (see 3.1.5.6)";
				break;
				case 5:
					message="Transparent Data Channels (32 channels) or ODA (see 3.1.5.8)";
				break;
				case 6:
					message="In House applications or ODA (see 3.1.5.9)";
				break;
				case 7:
					message="Radio Paging or ODA (see 3.1.5.10 and annex M)";
				break;
				case 8:
					message="Traffic Message Channel or ODA (see 3.1.5.12)";
				break;
				case 9:
					message="Emergency Warning System or ODA (see 3.1.5.13)";
				break;
				case 10:
					message="Programme Type Name";
				break;
				case 11:
					message="Open Data Applications";
				break;
				case 12:
					message="Open Data Applications";
				break;
				case 13:
					message="Enhanced Radio Paging or ODA (see annex M)";
				break;
				case 14:
					message="Enhanced Other Networks information only (see 3.1.5.19)";
				break;
				case 15:
					message="Defined in RBDS only";
				break;																																											
				default:
					printf("DECODE ERROR!!! (group_tye_code=%u)\n",group_type_code);
			}
			break;
		case 1:
			switch (group_type_code) {
				case 0:
					message="Basic tuning and switching information only (see 3.1.5.1)";
					do_type0(group,false);
				break;
				case 1:
					message="Programme Item Number (see 3.1.5.2)";
				break;
				case 2:
					message="RadioText only (see 3.1.5.3)";
					do_radiotext(group,false);
				break;
				case 3:
					message="Open Data Applications";
				break;
				case 4:
					message="Open Data Applications";
				break;
				case 5:
					message="Transparent Data Channels (32 channels) or ODA (see 3.1.5.8)";
				break;
				case 6:
					message="In House applications or ODA (see 3.1.5.9)";
				break;
				case 7:
					message="Open Data Applications";
				break;
				case 8:
					message="Open Data Applications";
				break;
				case 9:
					message="Open Data Applications";
				break;
				case 10:
					message="Open Data Applications";
				break;
				case 11:
					message="Open Data Applications";
				break;
				case 12:
					message="Open Data Applications";
				break;
				case 13:
					message="Open Data Applications";
				break;
				case 14:
					message="Enhanced Other Networks information only (see 3.1.5.19)";
				break;
				case 15:
					message="Fast switching information only (see 3.1.5.20)";
				break;																																											
				default:
					DBG(printf("DECODE ERROR!!! (group_tye_code=%u)\n",group_type_code);)
				break;
			}		
			break;
		default:
			DBG(printf("DECODE ERROR!!! (version_code=%u)\n",version_code);)
		break;
	}
	DBG(std::cout << "Group Type: " << message << std::endl;)
}

void gr_rds_data_decoder::decode_group(unsigned int group[]) {
	decode_pi(group[0]);
	decode_group_by_type((group[1]>>12) & 0x0f,(group[1]>>11) & 0x01,group);	
}

int gr_rds_data_decoder::work (int noutput_items,
			       gr_vector_const_void_star &input_items,
			       gr_vector_void_star &output_items)
{
  const bool *in = (const bool *) input_items[0];

  DBG(printf("RDS data decoder at work: input_items = %d, /104 = %d\n", noutput_items, noutput_items/104);)
  int i=0,j;
  unsigned long bit_distance,block_distance;
  unsigned int block_calculated_crc,block_received_crc,checkword,dataword;
	
	while (i<noutput_items) {
		reg=(reg<<1)|in[i];
		switch (d_state) {
			case ST_NO_SYNC:
				for (j=0;j<5;j++) {
					if (calc_syndrome(reg,26,0x5b9,10)==syndrome[j]) {
						DBG(printf("Offset %s syndrome found @ bit %lu\n",offset_name[j],bit_counter);)
						if (!presync) {
							lastseen_offset=j;
							lastseen_offset_counter=bit_counter;
							DBG(printf("out of presync %d %d %d\n",lastseen_offset,lastseen_offset_counter,bit_counter);)
							presync=true;
						}
						else {
							bit_distance=bit_counter-lastseen_offset_counter;
							if (offset_pos[lastseen_offset]>=offset_pos[j]) block_distance=offset_pos[j]+4-offset_pos[lastseen_offset];
							else block_distance=offset_pos[j]-offset_pos[lastseen_offset];
							DBG(printf("%lu,%lu\n",block_distance,bit_distance);)
							if ((block_distance*26)!=bit_distance) presync=false;
							else {
								printf("Sync State Detected\n");
								enter_sync(j);
							}	
						}
					j=5; //syndrome found, no more cycles						
					}
				}		
			break;
			case ST_SYNC:
				if (block_bit_counter<25) {
					block_bit_counter++;
				}
				else {
					good_block=false;
					DBG(printf("block_number:%u\n",block_number);)
					DBG(printf("reg:\t\t\t");printbin(reg,32);)
					//we got a complete block, let's see if CRC is correct
					dataword=(reg>>10) & 0xffff;
					DBG(printf("dataword:\t\t\t");printbin(dataword,16);)
					block_calculated_crc=calc_syndrome(dataword,16,0x5b9,10);	//calc CRC of the payload (16bits)
					DBG(printf("block_calculated_crc:\t\t");printbin(block_calculated_crc,10);)
					checkword=reg & 0x3ff;
					DBG(printf("checkword:\t\t\t");printbin(checkword,10);)
					if (block_number==2) {	//manage special case of C or C' offset word
						DBG(printf("offset_word:\t\t\t");printbin(offset_word[block_number],10);)
						block_received_crc=checkword^offset_word[block_number];
						DBG(printf("block_received_crc:\t\t");printbin(block_received_crc,10);)
						if (block_received_crc==block_calculated_crc) {
							DBG(printf("BLOCK RECEIVED OK\n");)
							good_block=true;
						}
						else {
							DBG(printf("C_ OFFSET WORD TEST\n");)
							DBG(printf("offset_word:\t\t\t");printbin(offset_word[4],10);)
							block_received_crc=checkword^offset_word[4];
							DBG(printf("block_received_crc:\t\t");printbin(block_received_crc,10);)
							if (block_received_crc==block_calculated_crc) {
								DBG(printf("BLOCK RECEIVED OK\n");)
								good_block=true;
							}
							else {			
								DBG(printf("BLOCK RECEIVED KO\n");)
								wrong_blocks_counter++;
								good_block=false;
							}
						}									
					}
					else {
						DBG(printf("offset_word:\t\t\t");printbin(offset_word[block_number],10);)
						block_received_crc=checkword^offset_word[block_number];
						DBG(printf("block_received_crc:\t\t");printbin(block_received_crc,10);)
						if (block_received_crc==block_calculated_crc) {
							DBG(printf("BLOCK RECEIVED OK\n");)
							good_block=true;
						}
						else {
							DBG(printf("BLOCK RECEIVED KO\n");)
							wrong_blocks_counter++;
							good_block=false;					
						}
					}
					if (block_number==0 && good_block) {
						group_assembly_started=true;
						group_good_blocks_counter=1;
						DBG(printf("Group Assembly Started @ Block %u\n",block_number);)						
					}
					if (group_assembly_started) {
						if (!good_block) {
							group_assembly_started=false;
							DBG(printf("Group Assembly Interrupted @ Block %u\n",block_number);)
						}
						else {
							group[block_number]=dataword;
							DBG(printf("Block Number %u Inserted in Group\n",block_number);)
							group_good_blocks_counter++;
						}
						if (group_good_blocks_counter==5) {
							DBG(printf("GROUP RECEIVED OK!!! (%04X %04X %04X %04X)\n",group[0],group[1],group[2],group[3]);)
							decode_group(group);
						}
					}
					block_bit_counter=0;				
					block_number=(block_number+1) % 4;
					blocks_counter++;
					if (blocks_counter==50) {
						if (wrong_blocks_counter>35) {
							printf("Lost Sync (Got %u bad blocks on %u total)\n",wrong_blocks_counter,blocks_counter);
							enter_no_sync();
						}
						blocks_counter=0;
						wrong_blocks_counter=0;
					}
				}
			break;
			default:
				d_state=ST_NO_SYNC;
			break;
		}
		i++;
		bit_counter++;
	}
	return noutput_items;
}
